// SPDX-License-Identifier: (BSD-3-Clause OR GPL-2.0-only)
/* Copyright(c) 2021 Intel Corporation */

#include <linux/debugfs.h>

#include "qat_dbg.h"
#include "qat_dbg_sysfs_cfg.h"
#include "adf_accel_devices.h"
#include "adf_cfg_common.h"
#include "qat_dbg_cfg.h"

#define QATD_SYS_BUFFER_SIZE 256
#define QATD_PARAM_MAX_SZ 128
#define QATD_PARAM_BASE_DEC 10
#define QATD_SYS_FS_PERM 0400

struct qatd_instance_config config_tbl[ADF_MAX_DEVICES];

static int qat_dbg_cfg_open(struct inode *inode, struct file *file)
{
	file->private_data = inode->i_private;

	return 0;
}

/**
 * qat_dbg_check_config() - Check debuggability sysfs configuration
 * @d_config: Pointer to debuggability configuration instance.
 * @accel_dev: Pointer to acceleration device.
 *
 * Function checks correctness of debugability parameters provided via sysfs.
 *
 * Return: 0 on success, error code otherwise
 */
static int qat_dbg_check_config(struct qatd_instance_config *d_config,
				struct adf_accel_dev *accel_dev)
{
	if (d_config->buffer_pool_size < QATD_BUF_POOL_MIN ||
	    d_config->buffer_pool_size > QATD_BUF_POOL_MAX) {
		dev_err(&GET_DEV(accel_dev),
			"Buffer pool should be in range %d..%d\n",
			QATD_BUF_POOL_MIN, QATD_BUF_POOL_MAX);
		return -EFAULT;
	}

	if (d_config->buffer_size < (QATD_BUF_SIZE_MIN << QATD_1MB_SHIFT) ||
	    d_config->buffer_size > (QATD_BUF_SIZE_MAX << QATD_1MB_SHIFT)) {
		dev_err(&GET_DEV(accel_dev),
			"Buffer size should be in range %d..%d\n",
			QATD_BUF_SIZE_MIN, QATD_BUF_SIZE_MAX);
		return -EFAULT;
	}

	if (strnlen(d_config->dump_dir, sizeof(d_config->dump_dir)) == 0) {
		dev_err(&GET_DEV(accel_dev),
			"Dump directory should be configured\n");
		return -EFAULT;
	}

	if (d_config->debug_level < 0 || d_config->debug_level > 3) {
		dev_err(&GET_DEV(accel_dev),
			"Debug level should be in range 0..3\n");
		return -EFAULT;
	}

	if (d_config->dump_on_proc_crash < 0 ||
	    d_config->dump_on_proc_crash > 1) {
		dev_err(&GET_DEV(accel_dev),
			"Dump on process crash should be in range 0..1\n");
		return -EFAULT;
	}

	if (d_config->sync_mode == QATD_SYNC_ON_CRASH)
		goto success;

	if (strnlen(d_config->cont_sync_dir,
		    sizeof(d_config->cont_sync_dir)) == 0) {
		dev_err(&GET_DEV(accel_dev),
			"Cont sync directory should be configured\n");
		return -EFAULT;
	}

	if (d_config->cont_sync_max_files_no < QATD_SYNC_FILES_MIN ||
	    d_config->cont_sync_max_files_no > QATD_SYNC_FILES_MAX) {
		dev_err(&GET_DEV(accel_dev),
			"Cont sync max files should be in range %d..%d\n",
			QATD_SYNC_FILES_MIN, QATD_SYNC_FILES_MAX);
		return -EFAULT;
	}

	if ((d_config->cont_sync_max_file_size <
	    (QATD_SYNC_FILE_SZ_MIN << QATD_1MB_SHIFT)) ||
	    (d_config->cont_sync_max_file_size >
	    (QATD_SYNC_FILE_SZ_MAX << QATD_1MB_SHIFT))) {
		dev_err(&GET_DEV(accel_dev),
			"Cont sync max files size should be in range %d..%d\n",
			QATD_SYNC_FILE_SZ_MIN, QATD_SYNC_FILE_SZ_MAX);
		return -EFAULT;
	}
success:

	return 0;
}

/**
 * qat_dbg_param_write() - Write debuggability parameters
 * @file: Pointer to a configuration file.
 * @user_buf: Pointer to user buffer.
 * @count: Maximum number of bytes to write.
 * @ppos: Current position in the buffer.
 *
 * Function provides mechanism to write debuggability parameters via sysfs.
 *
 * Return: Number of bytes wrote on success, error code otherwise
 */
static ssize_t qat_dbg_param_write(struct file *file,
				   const char __user *user_buffer,
				   size_t count, loff_t *position)
{
	struct adf_accel_dev *accel_dev;
	char str[QATD_SYS_BUFFER_SIZE] = {0};
	struct qatd_instance_config *d_config;
	int ret = 0, dbg_enabled = 0, cs_enabled = 0, reload_config = 0;
	char *d_iname = NULL;

	/* Basic defensive checks */
	if (!file)
		return -EFAULT;

	if (count == 0 || count > QATD_SYS_BUFFER_SIZE)
		return -EFAULT;

	ret = simple_write_to_buffer(str, QATD_SYS_BUFFER_SIZE,
				     position, user_buffer, count);

	if (ret != count)
		return -EFAULT;

	accel_dev = file->private_data;

	if (!accel_dev)
		return -EFAULT;

	d_config = &config_tbl[accel_dev->accel_id];

	if (file->f_path.dentry->d_iname)
		d_iname = file->f_path.dentry->d_iname;
	if (!d_iname)
		return -EFAULT;

	if (!strncmp(d_iname, QATD_PARAM_ENABLED, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC, &dbg_enabled))
			return -EFAULT;
		reload_config = 1;
	}
	if (!strncmp(d_iname, QATD_PARAM_DUMP_DIR, QATD_PARAM_MAX_SZ)) {
		memset(d_config->dump_dir, 0, sizeof(d_config->dump_dir));
		memcpy(d_config->dump_dir, str, count - 1);
	}
	if (!strncmp(d_iname, QATD_PARAM_DUMP_DIR_SZ, QATD_PARAM_MAX_SZ)) {
		if (kstrtoul(str, ADF_CFG_BASE_DEC,
			     &d_config->dump_dir_max_size))
			return -EFAULT;
		d_config->dump_dir_max_size <<= QATD_1MB_SHIFT;
	}
	if (!strncmp(d_iname, QATD_PARAM_BUFFER_POOL_SZ, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC,
			       &d_config->buffer_pool_size))
			return -EFAULT;
	}
	if (!strncmp(d_iname, QATD_PARAM_BUFFER_SZ, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC,
			       &d_config->buffer_size))
			return -EFAULT;
		d_config->buffer_size <<= QATD_1MB_SHIFT;
	}
	if (!strncmp(d_iname, QATD_PARAM_LEVEL, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC,
			       &d_config->debug_level))
			return -EFAULT;
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_ENABLED, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC, &cs_enabled))
			return -EFAULT;
		if (cs_enabled)
			d_config->sync_mode = QATD_SYNC_CONT;
		else
			d_config->sync_mode = QATD_SYNC_ON_CRASH;
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_DIR, QATD_PARAM_MAX_SZ)) {
		memset(d_config->cont_sync_dir, 0,
		       sizeof(d_config->cont_sync_dir));
		memcpy(d_config->cont_sync_dir, str, count - 1);
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_MAX_FILE_SZ, QATD_PARAM_MAX_SZ)) {
		if (kstrtoul(str, ADF_CFG_BASE_DEC,
			     &d_config->cont_sync_max_file_size))
			return -EFAULT;
		d_config->cont_sync_max_file_size <<= QATD_1MB_SHIFT;
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_MAX_FILES_NO, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC,
			       &d_config->cont_sync_max_files_no))
			return -EFAULT;
	}
	if (!strncmp(d_iname, QATD_PARAM_DUMP_ON_P_CRASH, QATD_PARAM_MAX_SZ)) {
		if (kstrtouint(str, ADF_CFG_BASE_DEC,
			       &d_config->dump_on_proc_crash))
			return -EFAULT;
	}

	if (!reload_config)
		return count;

	/* Attempt to turn on or reconfigure feature */
	if (dbg_enabled) {
		/* Check current config */
		ret = qat_dbg_check_config(d_config, accel_dev);
		if (ret)
			return -EFAULT;

		/* Instance exists - attempt to reconfigure */
		qat_dbg_shutdown_instance(accel_dev);

		ret = qat_dbg_init_instance_sysfs(accel_dev, d_config);
		if (ret)
			return -EFAULT;
		return count;
	}

	/* Disabling feature */
	qat_dbg_shutdown_instance(accel_dev);

	return count;
}

/**
 * qat_dbg_param_read() - Read debuggability parameters via sysfs
 * @file: Pointer to a configuration file.
 * @user_buf: Pointer to user buffer.
 * @count: Maximum number of bytes to read.
 * @ppos: Current position in the buffer.
 *
 * Function provides access to read debuggability parameters via sysfs.
 *
 * Return: Number of bytes read on success, error code or negative value
 * otherwise
 */
static ssize_t qat_dbg_param_read(struct file *file,
				  char __user *user_buf,
				  size_t count, loff_t *ppos)
{
	struct adf_accel_dev *accel_dev;
	char buf[QATD_SYS_BUFFER_SIZE];
	int len = 0;
	struct qatd_instance *dbg_inst;
	char *d_iname = NULL;
	int enabled = 0;

	if (!file)
		return -EFAULT;

	if (file->f_path.dentry->d_iname)
		d_iname = file->f_path.dentry->d_iname;

	if (!d_iname)
		return -EFAULT;

	accel_dev = file->private_data;

	if (!accel_dev)
		return -EFAULT;

	dbg_inst = accel_dev->qatd_instance;
	if (!strncmp(d_iname, QATD_PARAM_ENABLED, QATD_PARAM_MAX_SZ))
	{
		if (dbg_inst)
		{
			enabled = 1;
		} else
		{
			/* This case is hit while configuration is checked by
			 * the Debuggability Daemon. On particular device
			 * disabled, the Daemon needs to read whether
			 * Debuggability is enabled.
			 */
			enabled = qat_dbg_is_enabled(accel_dev);
			if (-ENODEV == enabled)
				return -EFAULT;
			enabled = !enabled;
		}

		len = scnprintf(buf, sizeof(buf), "%u\n", enabled);
		goto return_response;
	}

	if (!dbg_inst)
		return -EFAULT;

	if (!strncmp(d_iname, QATD_PARAM_DUMP_DIR, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%s\n",
				dbg_inst->config.dump_dir);
	}
	if (!strncmp(d_iname, QATD_PARAM_DUMP_DIR_SZ, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%lu\n",
				dbg_inst->config.dump_dir_max_size
				>> QATD_1MB_SHIFT);
	}
	if (!strncmp(d_iname, QATD_PARAM_BUFFER_POOL_SZ, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%u\n",
				dbg_inst->config.buffer_pool_size);
	}
	if (!strncmp(d_iname, QATD_PARAM_BUFFER_SZ, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%u\n",
				dbg_inst->config.buffer_size >> QATD_1MB_SHIFT);
	}
	if (!strncmp(d_iname, QATD_PARAM_LEVEL, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%u\n",
				dbg_inst->config.debug_level);
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_ENABLED, QATD_PARAM_MAX_SZ)) {
		if (dbg_inst->config.sync_mode == QATD_SYNC_CONT)
			len = scnprintf(buf, sizeof(buf), "%u\n", 1);
		else
			len = scnprintf(buf, sizeof(buf), "%u\n", 0);
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_DIR, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%s\n",
				dbg_inst->config.cont_sync_dir);
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_MAX_FILE_SZ, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%lu\n",
				dbg_inst->config.cont_sync_max_file_size >>
				QATD_1MB_SHIFT);
	}
	if (!strncmp(d_iname, QATD_PARAM_CS_MAX_FILES_NO, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%u\n",
				dbg_inst->config.cont_sync_max_files_no);
	}
	if (!strncmp(d_iname, QATD_PARAM_DUMP_ON_P_CRASH, QATD_PARAM_MAX_SZ)) {
		len = scnprintf(buf, sizeof(buf), "%u\n",
				dbg_inst->config.dump_on_proc_crash);
	}

return_response:
	if (len <= 0)
		return -EFAULT;

	return simple_read_from_buffer(user_buf, count, ppos, buf, len + 1);
}

static const struct file_operations qat_dbg_p_fops = {
	.open = qat_dbg_cfg_open,
	.read = qat_dbg_param_read,
	.write = qat_dbg_param_write
};

/**
 * qat_dbg_sysfs_cfg_add() - Add debuggability sysfs configuration files
 * @accel_dev: Pointer to acceleration device.
 *
 * Function creates debuggability files to expose debuggability parameters in
 * debugfs filesystem, in appropriate directory.
 *
 * Return: 0 on success, error code otherwise
 */
int qat_dbg_sysfs_cfg_add(struct adf_accel_dev *accel_dev)
{
	struct qatd_dentry_config *dentry_config;

	if (!accel_dev)
		return -EFAULT;

	accel_dev->qatd_config =
		kzalloc(sizeof(struct qatd_dentry_config), GFP_KERNEL);

	if (!accel_dev->qatd_config)
		return -EFAULT;

	dentry_config = accel_dev->qatd_config;

	/* Create directory for qat dbg */
	dentry_config->qat_dbg_dir =
		debugfs_create_dir(QATD_PARAM_SYSFS_DIR,
				   accel_dev->debugfs_dir);
	if (!dentry_config->qat_dbg_dir)
		goto err;

	/* Create files to expose QAT debug params */
	dentry_config->debug_enabled =
		debugfs_create_file(QATD_PARAM_ENABLED, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->debug_enabled)
		goto err;

	dentry_config->debug_level =
		debugfs_create_file(QATD_PARAM_LEVEL, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->debug_level)
		goto err;

	dentry_config->buffer_pool_size =
		debugfs_create_file(QATD_PARAM_BUFFER_POOL_SZ, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->buffer_pool_size)
		goto err;

	dentry_config->buffer_size_mb =
		debugfs_create_file(QATD_PARAM_BUFFER_SZ, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->buffer_size_mb)
		goto err;

	dentry_config->dump_dir =
		debugfs_create_file(QATD_PARAM_DUMP_DIR, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->dump_dir)
		goto err;

	dentry_config->dump_dir_size_mb =
		debugfs_create_file(QATD_PARAM_DUMP_DIR_SZ, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->dump_dir_size_mb)
		goto err;

	dentry_config->cont_sync_enabled =
		debugfs_create_file(QATD_PARAM_CS_ENABLED, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->cont_sync_enabled)
		goto err;

	dentry_config->cont_sync_log_dir =
		debugfs_create_file(QATD_PARAM_CS_DIR, QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->cont_sync_log_dir)
		goto err;

	dentry_config->cont_sync_max_log_files =
		debugfs_create_file(QATD_PARAM_CS_MAX_FILE_SZ,
				    QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->cont_sync_max_log_files)
		goto err;

	dentry_config->cont_sync_max_log_size_mb =
		debugfs_create_file(QATD_PARAM_CS_MAX_FILES_NO,
				    QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->cont_sync_max_log_size_mb)
		goto err;

	dentry_config->dump_on_crash =
		debugfs_create_file(QATD_PARAM_DUMP_ON_P_CRASH,
				    QATD_SYS_FS_PERM,
				    dentry_config->qat_dbg_dir,
				    accel_dev,
				    &qat_dbg_p_fops);
	if (!dentry_config->dump_on_crash)
		goto err;

	return 0;

err:
	dev_err(&GET_DEV(accel_dev),
		"Failed to create QAT dbg debugfs entry.\n");
	qat_dbg_sysfs_cfg_del(accel_dev);

	return -EFAULT;
}

void qat_dbg_sysfs_cfg_del(struct adf_accel_dev *accel_dev)
{
	struct qatd_dentry_config *dentry_config;

	if (!accel_dev)
		return;

	dentry_config = accel_dev->qatd_config;
	if (!dentry_config)
		return;

	debugfs_remove(dentry_config->debug_enabled);
	dentry_config->debug_enabled = NULL;

	debugfs_remove(dentry_config->debug_level);
	dentry_config->debug_level = NULL;

	debugfs_remove(dentry_config->buffer_pool_size);
	dentry_config->buffer_pool_size = NULL;

	debugfs_remove(dentry_config->buffer_size_mb);
	dentry_config->buffer_size_mb = NULL;

	debugfs_remove(dentry_config->dump_dir);
	dentry_config->dump_dir = NULL;

	debugfs_remove(dentry_config->dump_dir_size_mb);
	dentry_config->dump_dir_size_mb = NULL;

	debugfs_remove(dentry_config->cont_sync_enabled);
	dentry_config->cont_sync_enabled = NULL;

	debugfs_remove(dentry_config->cont_sync_log_dir);
	dentry_config->cont_sync_log_dir = NULL;

	debugfs_remove(dentry_config->cont_sync_max_log_files);
	dentry_config->cont_sync_max_log_files = NULL;

	debugfs_remove(dentry_config->cont_sync_max_log_size_mb);
	dentry_config->cont_sync_max_log_size_mb = NULL;

	debugfs_remove(dentry_config->dump_on_crash);
	dentry_config->dump_on_crash = NULL;

	debugfs_remove(dentry_config->qat_dbg_dir);
	dentry_config->qat_dbg_dir = NULL;

	kfree(accel_dev->qatd_config);
	accel_dev->qatd_config = NULL;
}
